# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Quan Zhou <quan@bitergia.com>
#


import logging
import unittest

import requests
from base import TestBaseBackend
from grimoire_elk.enriched.utils import REPO_LABELS


class TestWeblate(TestBaseBackend):
    """Test Weblate backend"""

    connector = "weblate"
    ocean_index = "test_" + connector
    enrich_index = "test_" + connector + "_enrich"

    def test_has_identites(self):
        """Test value of has_identities method"""

        enrich_backend = self.connectors[self.connector][2]()
        self.assertTrue(enrich_backend.has_identities())

    def test_items_to_raw(self):
        """Test whether JSON items are properly inserted into ES"""

        result = self._test_items_to_raw()
        self.assertEqual(result['items'], 4)
        self.assertEqual(result['raw'], 4)

        author_data = [
            {
                "email": "quan@bitergia.com",
                "full_name": "quan zhou",
                "username": "quan",
                "groups": [
                    "https://my.weblate.org/api/groups/2/",
                    "https://my.weblate.org/api/groups/3/"
                ],
                "is_superuser": False,
                "is_active": True,
                "date_joined": "2020-09-16T09:03:33.581180Z",
                "url": "https://my.weblate.org/api/users/1/"
            },
            {
                "email": "tom@jerry.com",
                "full_name": "tom jerry",
                "username": "tom",
                "groups": [
                    "https://my.weblate.org/api/groups/2/",
                    "https://my.weblate.org/api/groups/3/"
                ],
                "is_superuser": False,
                "is_active": True,
                "date_joined": "2020-09-17T09:03:33.581180Z",
                "url": "https://my.weblate.org/api/users/2/"
            },
            {
                "email": "tom@jerry.com",
                "full_name": "tom jerry",
                "username": "tom",
                "groups": [
                    "https://my.weblate.org/api/groups/2/",
                    "https://my.weblate.org/api/groups/3/"
                ],
                "is_superuser": False,
                "is_active": True,
                "date_joined": "2020-09-17T09:03:33.581180Z",
                "url": "https://my.weblate.org/api/users/2/"
            }
        ]
        for i, item in enumerate(self.items):
            self.ocean_backend._fix_item(item)

            if i == 3:
                self.assertNotIn('author_data', item['data'])
                self.assertIsNone(item['data']['author'])
            else:
                self.assertEqual(author_data[i], item['data']['author_data'])

    def test_raw_to_enrich(self):
        """Test whether the raw index is properly enriched"""

        result = self._test_raw_to_enrich()
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

        enrich_backend = self.connectors[self.connector][2]()

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_full_name'], 'quan zhou')
        self.assertEqual(eitem['change_id'], 1)
        self.assertEqual(eitem['change_api_url'], 'https://my.weblate.org/api/changes/1/')
        self.assertIsNone(eitem['component_api_url'])
        self.assertIsNone(eitem['translation_api_url'])
        self.assertEqual(eitem['unit_id'], 1)
        self.assertEqual(eitem['unit_context'], 'content.text')
        self.assertEqual(eitem['unit_api_url'], 'https://my.weblate.org/api/units/1')

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_full_name'], 'tom jerry')
        self.assertEqual(eitem['change_id'], 2)
        self.assertEqual(eitem['change_api_url'], 'https://my.weblate.org/api/changes/2/')
        self.assertIsNone(eitem['component_api_url'])
        self.assertIsNone(eitem['translation_api_url'])
        self.assertNotIn('unit_context', eitem)

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_full_name'], 'tom jerry')
        self.assertEqual(eitem['change_id'], 3)
        self.assertEqual(eitem['change_api_url'], 'https://my.weblate.org/api/changes/3/')
        self.assertEqual(eitem['component_api_url'], 'https://my.weblate.org/api/components/master/native/')
        self.assertEqual(eitem['component_name'], 'native')
        self.assertEqual(eitem['translation_api_url'], 'https://my.weblate.org/api/translations/master/native/es')
        self.assertEqual(eitem['translation_name'], 'es')
        self.assertEqual(eitem['project_name'], 'master')
        self.assertNotIn('unit_context', eitem)

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertNotIn('author_full_name', eitem)
        self.assertEqual(eitem['change_id'], 4)
        self.assertEqual(eitem['change_api_url'], 'https://my.weblate.org/api/changes/4/')
        self.assertEqual(eitem['component_api_url'], 'https://my.weblate.org/api/components/italian/nativo/')
        self.assertEqual(eitem['component_name'], 'nativo')
        self.assertEqual(eitem['translation_api_url'], 'https://my.weblate.org/api/translations/italian/nativo/it')
        self.assertEqual(eitem['translation_name'], 'it')
        self.assertEqual(eitem['project_name'], 'italian')
        self.assertNotIn('unit_context', eitem)

    def test_enrich_repo_labels(self):
        """Test whether the field REPO_LABELS is present in the enriched items"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            self.assertIn(REPO_LABELS, eitem)

    def test_raw_to_enrich_sorting_hat(self):
        """Test enrich with SortingHat"""

        result = self._test_raw_to_enrich(sortinghat=True)
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

        enrich_backend = self.connectors[self.connector][2]()
        enrich_backend.sortinghat = True

        item = self.items[0]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_name'], 'quan zhou')
        self.assertEqual(eitem['author_user_name'], 'quan')
        self.assertEqual(eitem['author_org_name'], 'Unknown')
        self.assertEqual(eitem['author_multi_org_names'], ['Unknown'])

        item = self.items[1]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_name'], 'tom jerry')
        self.assertEqual(eitem['author_user_name'], 'tom')
        self.assertEqual(eitem['author_org_name'], 'Unknown')
        self.assertEqual(eitem['author_multi_org_names'], ['Unknown'])

        item = self.items[2]
        eitem = enrich_backend.get_rich_item(item)
        self.assertEqual(eitem['author_name'], 'tom jerry')
        self.assertEqual(eitem['author_user_name'], 'tom')
        self.assertEqual(eitem['author_org_name'], 'Unknown')
        self.assertEqual(eitem['author_multi_org_names'], ['Unknown'])

        item = self.items[3]
        eitem = enrich_backend.get_rich_item(item)
        self.assertNotIn('author_name', eitem)
        self.assertNotIn('author_user_name', eitem)
        self.assertNotIn('author_org_name', eitem)
        self.assertNotIn('author_multi_org_names', eitem)

    def test_raw_to_enrich_projects(self):
        """Test enrich with Projects"""

        result = self._test_raw_to_enrich(projects=True)
        self.assertEqual(result['raw'], 4)
        self.assertEqual(result['enrich'], 4)

        res = requests.get(self.es_con + "/" + self.enrich_index + "/_search", verify=False)
        for eitem in res.json()['hits']['hits']:
            self.assertEqual(eitem['_source']['project'], "grimoire")

    def test_copy_raw_fields(self):
        """Test copied raw fields"""

        self._test_raw_to_enrich()
        enrich_backend = self.connectors[self.connector][2]()

        for item in self.items:
            eitem = enrich_backend.get_rich_item(item)
            for attribute in enrich_backend.RAW_FIELDS_COPY:
                if attribute in item:
                    self.assertEqual(item[attribute], eitem[attribute])
                else:
                    self.assertIsNone(eitem[attribute])


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format='%(asctime)s %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    unittest.main(warnings='ignore')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Animesh Kumar<animuz111@gmail.com>
#   Obaro Ikoh <obaroikohb@gmail.com>
#

import logging

from grimoirelab_toolkit.datetime import str_to_datetime

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "msg_analyzed": {
                  "type": "text",
                  "fielddata": true,
                  "index": true
                }
           }
        } """

        return {"items": mapping}


class RocketChatEnrich(Enrich):
    mapping = Mapping

    def get_field_author(self):
        return "u"

    def get_sh_identity(self, item, identity_field=None):
        identity = {
            'username': None,
            'name': None,
            'email': None
        }

        if self.get_field_author() not in item['data']:
            return identity
        user = item['data'][self.get_field_author()]

        identity['username'] = user.get('username', None)
        identity['name'] = user.get('name', None)

        return identity

    def get_identities(self, item):
        """ Return the identities from an item """

        identity = self.get_sh_identity(item)
        yield identity

    def get_project_repository(self, eitem):
        tokens = eitem['origin'].rsplit("/", 1)
        return tokens[0] + " " + tokens[1]

    @metadata
    def get_rich_item(self, item):

        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        message = item['data']

        eitem['msg_analyzed'] = message['msg']
        eitem['msg'] = message['msg']
        eitem['rid'] = message['rid']
        eitem['msg_id'] = message['_id']
        # parent exists in case message is a reply
        eitem['msg_parent'] = message.get('parent', None)

        author = message.get('u', None)
        if author:
            eitem['user_id'] = author.get('_id', None)
            eitem['user_name'] = author.get('name', None)
            eitem['user_username'] = author.get('username', None)

        eitem['is_edited'] = 0
        editor = message.get('editedBy', None)
        if editor:
            eitem['edited_at'] = str_to_datetime(message['editedAt']).isoformat()
            eitem['edited_by_username'] = editor.get('username', None)
            eitem['edited_by_user_id'] = editor.get('_id', None)
            eitem['is_edited'] = 1

        file = message.get('file', None)
        if file:
            eitem['file_id'] = file.get('_id', None)
            eitem['file_name'] = file.get('name', None)
            eitem['file_type'] = file.get('type', None)

        eitem['replies'] = len(message['replies']) if message.get('replies', None) else 0

        eitem['total_reactions'] = 0
        reactions = message.get('reactions', None)
        if reactions:
            reaction_types, total_reactions = self.__get_reactions(reactions)
            eitem.update({'reactions': reaction_types})
            eitem['total_reactions'] = total_reactions

        eitem['total_mentions'] = 0
        mentions = message.get('mentions', None)
        if mentions:
            eitem['mentions'] = self.__get_mentions(mentions)
            eitem['total_mentions'] = len(mentions)

        channel_info = message.get('channel_info', None)
        if channel_info:
            eitem.update(self.__get_channel_info(channel_info))

        eitem['total_urls'] = 0
        urls = message.get('urls', None)
        if urls:
            urls = [{'url': url['url']} for url in urls]
            eitem['message_urls'] = urls
            eitem['total_urls'] = len(urls)

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "message"))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def __get_reactions(self, reactions):
        """Enrich reactions for the message"""

        reaction_types = []
        total_reactions = 0
        for reaction_type in reactions:
            reaction_data = reactions[reaction_type]
            usernames = reaction_data.get('usernames', [])
            names = reaction_data.get('names', [])
            reaction_type = {
                "type": reaction_type,
                "username": usernames,
                "names": names,
                "count": len(usernames)
            }
            total_reactions += len(usernames)
            reaction_types.append(reaction_type)

        return reaction_types, total_reactions

    def __get_mentions(self, mentioned):
        """Enrich users mentioned in the message"""

        rich_mentions = []

        for usr in mentioned:
            rich_mention = {
                'username': usr.get('username', None),
                'id': usr.get('_id', None),
                'name': usr.get('name', None)
            }
            rich_mentions.append(rich_mention)

        return rich_mentions

    def __get_channel_info(self, channel):
        """Enrich channel info of the message"""

        rich_channel = {
            'channel_id': channel['_id'],
            'channel_updated_at': str_to_datetime(channel['_updatedAt']).isoformat(),
            'channel_num_messages': channel.get('msgs', None),
            'channel_name': channel.get('name', ''),
            'channel_num_users': channel.get('usersCount', None),
            'channel_topic': channel.get('topic', ''),
            'avatar': ''
        }
        if 'lastMessage' in channel and channel['lastMessage']:
            rich_channel['avatar'] = channel['lastMessage']['avatar']

        return rich_channel



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#     Valerio Cosentino <valcos@bitergia.com>
#

import configparser
import json
import os
import requests
import sys
import unittest
from datetime import datetime

from elasticsearch import Elasticsearch

if '..' not in sys.path:
    sys.path.insert(0, '..')

from grimoire_elk.elk import load_identities
from grimoire_elk.utils import get_connectors, get_elastic
from tests.model import ESMapping

CONFIG_FILE = 'tests.conf'
DB_SORTINGHAT = "test_sh"
DB_PROJECTS = "test_projects"
FILE_PROJECTS = "data/projects-release.json"
SCHEMA_DIR = '../schema/'


def load_mapping(enrich_index, csv_name):

    cvs_path = os.path.join(SCHEMA_DIR, csv_name + '.csv')
    cvs_mapping = ESMapping.from_csv(enrich_index, cvs_path)

    return cvs_mapping


def data2es(items, ocean):
    def ocean_item(item):
        # Hack until we decide when to drop this field
        if 'updated_on' in item:
            updated = datetime.fromtimestamp(item['updated_on'])
            item['metadata__updated_on'] = updated.isoformat()
        if 'timestamp' in item:
            ts = datetime.fromtimestamp(item['timestamp'])
            item['metadata__timestamp'] = ts.isoformat()

        # the _fix_item does not apply to the test data for Twitter
        try:
            ocean._fix_item(item)
        except KeyError:
            pass

        if ocean.anonymize:
            ocean.identities.anonymize_item(item)

        return item

    items_pack = []  # to feed item in packs

    for item in items:
        item = ocean_item(item)
        if len(items_pack) >= ocean.elastic.max_items_bulk:
            ocean._items_to_es(items_pack)
            items_pack = []
        items_pack.append(item)
    inserted = ocean._items_to_es(items_pack)

    return inserted


def refresh_identities(enrich_backend):
    total = 0

    for eitem in enrich_backend.fetch():
        roles = None
        try:
            roles = enrich_backend.roles
        except AttributeError:
            pass
        new_identities = enrich_backend.get_item_sh_from_id(eitem, roles)
        eitem.update(new_identities)
        total += 1

    return total


def refresh_projects(enrich_backend):
    total = 0

    for eitem in enrich_backend.fetch():
        new_project = enrich_backend.get_item_project(eitem)
        eitem.update(new_project)
        total += 1

    return total


class TestBaseBackend(unittest.TestCase):
    """Functional tests for GrimoireELK Backends"""

    @classmethod
    def setUpClass(cls):
        cls.config = configparser.ConfigParser()
        cls.config.read(CONFIG_FILE)
        cls.es_con = dict(cls.config.items('ElasticSearch'))['url']
        cls.connectors = get_connectors()
        cls.maxDiff = None

        # Sorting hat settings
        cls.db_user = ''
        cls.db_password = ''
        cls.db_projects_map = 'test_projects'
        if 'Database' in cls.config:
            if 'user' in cls.config['Database']:
                cls.db_user = cls.config['Database']['user']
            if 'password' in cls.config['Database']:
                cls.db_password = cls.config['Database']['password']

    def setUp(self):
        with open(os.path.join("data", self.connector + ".json")) as f:
            self.items = json.load(f)

        self.ocean_backend = None
        self.enrich_backend = None
        self.ocean_aliases = []
        self.enrich_aliases = []

    def tearDown(self):
        delete_test_idx = self.es_con + "/" + 'test*'
        requests.delete(delete_test_idx, verify=False)

    def _test_items_to_raw(self):
        """Test whether fetched items are properly loaded to ES"""

        clean = True
        perceval_backend = None
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index, clean, self.ocean_backend, self.ocean_aliases)
        self.ocean_backend.set_elastic(elastic_ocean)

        raw_items = data2es(self.items, self.ocean_backend)

        return {'items': len(self.items), 'raw': raw_items}

    def _test_raw_to_enrich(self, sortinghat=False, projects=False, pair_programming=False):
        """Test whether raw indexes are properly enriched"""

        # populate raw index
        perceval_backend = None
        clean = True
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index, clean, self.ocean_backend)
        self.ocean_backend.set_elastic(elastic_ocean)
        data2es(self.items, self.ocean_backend)

        # populate enriched index
        if not sortinghat and not projects:
            self.enrich_backend = self.connectors[self.connector][2]()
        elif sortinghat and not projects:
            self.enrich_backend = self.connectors[self.connector][2](db_sortinghat=DB_SORTINGHAT,
                                                                     db_user=self.db_user,
                                                                     db_password=self.db_password)
        elif not sortinghat and projects:
            self.enrich_backend = self.connectors[self.connector][2](json_projects_map=FILE_PROJECTS,
                                                                     db_user=self.db_user,
                                                                     db_password=self.db_password)
        if pair_programming:
            self.enrich_backend.pair_programming = pair_programming

        elastic_enrich = get_elastic(self.es_con, self.enrich_index, clean, self.enrich_backend, self.enrich_aliases)
        self.enrich_backend.set_elastic(elastic_enrich)

        # Load SH identities
        if sortinghat:
            load_identities(self.ocean_backend, self.enrich_backend)

        raw_count = len([item for item in self.ocean_backend.fetch()])
        enrich_count = self.enrich_backend.enrich_items(self.ocean_backend)
        # self._test_csv_mappings(sortinghat)

        return {'raw': raw_count, 'enrich': enrich_count}

    def _test_csv_mappings(self, sortinghat):
        """Test whether the mappings in the CSV are successfully met"""

        result = {}

        if not sortinghat:
            return result

        csv_mapping = load_mapping(self.enrich_index, self.connector)
        client = Elasticsearch(self.es_con, timeout=30)
        mapping_json = client.indices.get_mapping(index=self.enrich_index)
        es_mapping = ESMapping.from_json(index_name=self.enrich_index,
                                         mapping_json=mapping_json)

        result = csv_mapping.compare_properties(es_mapping)
        self.assertEqual(result['msg'], "")

    def _test_refresh_identities(self):
        """Test refresh identities"""

        # populate raw index
        perceval_backend = None
        clean = True
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index, clean, self.ocean_backend)
        self.ocean_backend.set_elastic(elastic_ocean)
        data2es(self.items, self.ocean_backend)

        # populate enriched index
        self.enrich_backend = self.connectors[self.connector][2]()
        load_identities(self.ocean_backend, self.enrich_backend)
        self.enrich_backend = self.connectors[self.connector][2](db_sortinghat=DB_SORTINGHAT,
                                                                 db_user=self.db_user,
                                                                 db_password=self.db_password)
        elastic_enrich = get_elastic(self.es_con, self.enrich_index, clean, self.enrich_backend)
        self.enrich_backend.set_elastic(elastic_enrich)
        self.enrich_backend.enrich_items(self.ocean_backend)

        total = refresh_identities(self.enrich_backend)
        return total

    def _test_refresh_project(self):
        """Test refresh project field"""

        # populate raw index
        perceval_backend = None
        clean = True
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index, clean, self.ocean_backend)
        self.ocean_backend.set_elastic(elastic_ocean)
        data2es(self.items, self.ocean_backend)

        # populate enriched index
        self.enrich_backend = self.connectors[self.connector][2](db_projects_map=DB_PROJECTS,
                                                                 db_user=self.db_user,
                                                                 db_password=self.db_password)

        elastic_enrich = get_elastic(self.es_con, self.enrich_index, clean, self.enrich_backend)
        self.enrich_backend.set_elastic(elastic_enrich)
        self.enrich_backend.enrich_items(self.ocean_backend)

        total = refresh_projects(self.enrich_backend)
        return total

    def _test_study(self, test_study, projects_json_repo=None, projects_json=None, prjs_map=None):
        """Test the execution of a study"""

        # populate raw index
        perceval_backend = None
        clean = True
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index, clean, self.ocean_backend)
        self.ocean_backend.set_elastic(elastic_ocean)
        data2es(self.items, self.ocean_backend)

        # populate enriched index
        self.enrich_backend = self.connectors[self.connector][2](db_sortinghat=DB_SORTINGHAT,
                                                                 db_user=self.db_user,
                                                                 db_password=self.db_password,
                                                                 db_projects_map=self.db_projects_map)

        elastic_enrich = get_elastic(self.es_con, self.enrich_index, clean, self.enrich_backend)
        self.enrich_backend.set_elastic(elastic_enrich)

        if projects_json:
            self.enrich_backend.json_projects = projects_json

        if projects_json_repo:
            self.enrich_backend.projects_json_repo = projects_json_repo

        if prjs_map:
            self.enrich_backend.prjs_map = prjs_map

        self.enrich_backend.enrich_items(self.ocean_backend)

        for study in self.enrich_backend.studies:
            if test_study == study.__name__:
                found = (study, self.ocean_backend, self.enrich_backend)
                break

        return found

    def _test_items_to_raw_anonymized(self):
        clean = True
        perceval_backend = None
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend, anonymize=True)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index_anonymized, clean, self.ocean_backend,
                                    self.ocean_aliases)
        self.ocean_backend.set_elastic(elastic_ocean)

        raw_items = data2es(self.items, self.ocean_backend)

        return {'items': len(self.items), 'raw': raw_items}

    def _test_raw_to_enrich_anonymized(self, sortinghat=False, projects=False):
        """Test whether raw indexes are properly enriched"""

        # populate raw index
        perceval_backend = None
        clean = True
        self.ocean_backend = self.connectors[self.connector][1](perceval_backend, anonymize=True)
        elastic_ocean = get_elastic(self.es_con, self.ocean_index_anonymized, clean, self.ocean_backend)
        self.ocean_backend.set_elastic(elastic_ocean)
        data2es(self.items, self.ocean_backend)

        # populate enriched index
        self.enrich_backend = self.connectors[self.connector][2]()

        elastic_enrich = get_elastic(self.es_con, self.enrich_index_anonymized, clean, self.enrich_backend, self.enrich_aliases)
        self.enrich_backend.set_elastic(elastic_enrich)

        raw_count = len([item for item in self.ocean_backend.fetch()])
        enrich_count = self.enrich_backend.enrich_items(self.ocean_backend)

        return {'raw': raw_count, 'enrich': enrich_count}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Nitish Gupta <imnitish.ng@gmail.com>
#

import logging

from grimoirelab_toolkit.datetime import (datetime_utcnow)

from .utils import get_time_diff_days

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        geopoints type is not created in dynamic mapping

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
               "description_analyzed": {
                    "type": "text",
                    "index": true
               }
           }
        }
        """
        return {"items": mapping}


class LaunchpadEnrich(Enrich):

    mapping = Mapping
    roles = ['assignee_data', 'owner_data']

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                 db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map,
                         db_user, db_password, db_host)

    def get_field_author(self):
        return "owner_data"

    def set_elastic(self, elastic):
        self.elastic = elastic

    def get_sh_identity(self, item, identity_field=None):
        """ Return a Sorting Hat identity using launchpad user data """

        identity = {
            'username': None,
            'name': None,
            'email': None
        }

        item = item['data']
        if isinstance(item, dict) and identity_field in item:
            identity['username'] = item[identity_field].get('name', None)
            identity['name'] = item[identity_field].get('display_name', None)

        return identity

    def get_identities(self, item):
        """ Return the identities from an item """

        for rol in self.roles:
            if rol in item['data']:
                user = self.get_sh_identity(item, rol)
                yield user

    @metadata
    def get_rich_item(self, item):

        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        # data fields to copy
        copy_fields = ["title", "web_link", "date_created", "date_incomplete", "is_complete",
                       "status", "bug_target_name", "importance", "date_triaged", "date_left_new"]

        bug = item['data']
        self.copy_raw_fields(copy_fields, bug, eitem)

        if self.sortinghat:
            eitem.update(self.get_item_sh(item, self.roles))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem.update(self.__get_rich_bugs(bug))
        eitem.update(self.get_grimoire_fields(bug["date_created"], "issue"))

        return eitem

    def __get_rich_bugs(self, data):
        """Create enriched data for bugs"""

        rich_bugtask = {}

        # Time to
        if not data["is_complete"]:
            rich_bugtask["time_open_days"] = get_time_diff_days(data['date_created'], datetime_utcnow().replace(tzinfo=None))
        else:
            rich_bugtask["time_open_days"] = get_time_diff_days(data['date_created'], data['date_closed'])
            rich_bugtask["time_created_to_assigned"] = get_time_diff_days(data['date_created'], data['date_assigned'])
            rich_bugtask['time_assigned_to_closed'] = get_time_diff_days(data['date_assigned'], data['date_closed'])
            rich_bugtask["time_to_close_days"] = get_time_diff_days(data['date_created'], data['date_closed'])

        if data['activity_data']:
            rich_bugtask['time_to_last_update_days'] = \
                get_time_diff_days(data['date_created'], data['activity_data'][-1]['datechanged'])

        rich_bugtask['reopened'] = 1 if data['date_left_closed'] else 0
        rich_bugtask['time_to_fix_commit'] = get_time_diff_days(data['date_created'], data['date_fix_committed'])
        rich_bugtask['time_worked_on'] = get_time_diff_days(data['date_in_progress'], data['date_fix_committed'])
        rich_bugtask['time_to_confirm'] = get_time_diff_days(data['date_created'], data['date_confirmed'])

        # Author and assignee data
        owner = data.get('owner_data', None)
        if owner:
            rich_bugtask['user_login'] = owner.get('name', None)
            rich_bugtask['user_name'] = owner.get('display_name', None)
            rich_bugtask['user_joined'] = owner.get('date_created', None)
            rich_bugtask['user_karma'] = owner.get('karma', None)
            rich_bugtask['user_time_zone'] = owner.get('time_zone', None)

        assignee = data.get('assignee_data', None)
        if assignee:
            assignee = data['assignee_data']
            rich_bugtask['assignee_login'] = assignee.get('name', None)
            rich_bugtask['assignee_name'] = assignee.get('display_name', None)
            rich_bugtask['assignee_joined'] = assignee.get('date_created', None)
            rich_bugtask['assignee_karma'] = assignee.get('karma', None)
            rich_bugtask['assignee_time_zone'] = assignee.get('time_zone', None)

        # Extract info related to bug
        rich_bugtask.update(self.__extract_bug_info(data['bug_data']))

        rich_bugtask['time_to_first_attention'] = \
            get_time_diff_days(data['date_created'], self.get_time_to_first_attention(data))
        rich_bugtask['activity_count'] = len(data['activity_data'])

        return rich_bugtask

    def __extract_bug_info(self, bug_data):

        rich_bug_info = {}

        copy_fields = ["latest_patch_uploaded", "security_related", "private", "users_affected_count",
                       "title", "description", "tags", "date_last_updated", "message_count", "heat"]
        rich_bug_info['time_created_to_last_update_days'] = \
            get_time_diff_days(bug_data['date_created'], bug_data['date_last_updated'])

        rich_bug_info['description'] = bug_data['description'][:self.KEYWORD_MAX_LENGTH]
        rich_bug_info['description_analyzed'] = bug_data['description']
        rich_bug_info['bug_name'] = bug_data['name']
        rich_bug_info['bug_id'] = bug_data['id']

        self.copy_raw_fields(copy_fields, bug_data, rich_bug_info)

        return rich_bug_info

    def get_time_to_first_attention(self, item):
        """Get the first date at which a comment or activity was made to the bug by someone
        other than the user who created the issue
        """
        message_dates = [message['date_created'] for message in item['messages_data']
                         if item['owner_data'].get('name', None) != message['owner_data'].get('name', None)]
        activity_dates = [activity['datechanged'] for activity in item['activity_data']
                          if item['owner_data'].get('name', None) != activity['person_data'].get('name', None)]
        activity_dates.extend(message_dates)
        if activity_dates:
            return min(activity_dates)
        return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from requests.structures import CaseInsensitiveDict

from .enrich import Enrich, metadata, DEFAULT_PROJECT
from ..elastic_mapping import Mapping as BaseMapping
from grimoirelab_toolkit.datetime import str_to_datetime

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "text_analyzed": {
                  "type": "text",
                  "index": true
                },
                "geolocation": {
                     "type": "geo_point"
                }
           }
        } """

        return {"items": mapping}


class TwitterEnrich(Enrich):

    mapping = Mapping

    def get_field_author(self):
        return "user"

    def get_sh_identity(self, item, identity_field=None):
        identity = {}
        identity['username'] = None
        identity['email'] = None
        identity['name'] = None

        if identity_field is None:
            identity_field = self.get_field_author()

        tweet = item  # by default a specific user dict is expected
        if isinstance(item, dict) and 'data' in item:
            tweet = item['data']

        if identity_field in tweet:
            identity['username'] = tweet[identity_field]['screen_name']
            identity['name'] = tweet[identity_field]['name']
        return identity

    def get_identities(self, item):
        """ Return the identities from an item """

        item = item['data']
        user = self.get_sh_identity(item)

        yield user

    def get_item_project(self, eitem):
        """ Get project mapping enrichment field.

        Twitter mappings is pretty special so it needs a special
        implementacion.
        """

        project = None
        eitem_project = {}
        ds_name = self.get_connector_name()  # data source name in projects map

        if ds_name not in self.prjs_map:
            return eitem_project

        for tag in eitem['hashtags_analyzed']:
            # lcanas: hashtag provided in projects.json file should not be case sensitive T6876
            tags2project = CaseInsensitiveDict(self.prjs_map[ds_name])
            if tag in tags2project:
                project = tags2project[tag]
                break

        if project is None:
            project = DEFAULT_PROJECT

        eitem_project = {"project": project}

        eitem_project.update(self.add_project_levels(project))

        return eitem_project

    @metadata
    def get_rich_item(self, item):
        eitem = {}
        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        # The real data
        tweet = item['data']

        # data fields to copy
        copy_fields = ["id", "id_str", "lang", "place", "retweet_count",
                       "text", "in_reply_to_user_id_str",
                       "in_reply_to_screen_name"]
        for f in copy_fields:
            if f in tweet:
                eitem[f] = tweet[f]
            else:
                eitem[f] = None

        # Date fields
        eitem["created_at"] = str_to_datetime(tweet["created_at"]).isoformat()

        # data fields to copy from user
        copy_fields = ["created_at", "description", "followers_count",
                       "friends_count", "id_str", "location", "name",
                       "url", "verified"]
        for f in copy_fields:
            if f in tweet['user']:
                eitem["user_" + f] = tweet['user'][f]
            else:
                eitem["user_" + f] = None

        if "text" in tweet:
            eitem["text_analyzed"] = tweet["text"]

        eitem['hashtags_analyzed'] = []
        for tag in tweet['entities']['hashtags']:
            eitem['hashtags_analyzed'].append(tag['text'])

        eitem['retweeted'] = 0
        if tweet['retweeted']:
            eitem['retweeted'] = 1

        eitem['url'] = "http://twitter.com/" + tweet['user']['screen_name']
        eitem['url'] += "/status/" + tweet['id_str']
        eitem['user_url_twitter'] = "http://twitter.com/" + tweet['user']['screen_name']

        eitem.update(self.get_grimoire_fields(item["metadata__updated_on"], "tweet"))

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from grimoirelab_toolkit.datetime import (str_to_datetime,
                                          unixtime_to_datetime)

from .utils import get_time_diff_days

from .enrich import Enrich, metadata
from ..elastic_mapping import Mapping as BaseMapping

MAX_SIZE_BULK_ENRICHED_ITEMS = 200


logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
                "author_badges": {
                  "type": "text",
                  "index": true
                },
                "summary": {
                  "type": "text",
                  "index": true
                },
                "id": {
                  "type": "keyword"
                }
           }
        } """

        return {"items": mapping}


class AskbotEnrich(Enrich):

    mapping = Mapping

    def get_identities(self, item):
        """ Return the identities from an item """

        # question
        user = self.get_sh_identity(item, self.get_field_author())
        yield user

        # answers
        if 'answers' in item['data']:
            for answer in item['data']['answers']:
                # avoid "answered_by" : "This post is a wiki" corner case
                if type(answer['answered_by']) is dict:
                    user = self.get_sh_identity(answer['answered_by'])
                    yield user
                if 'comments' in answer:
                    for comment in answer['comments']:
                        commenter = self.get_sh_identity(comment)
                        yield commenter

    def get_sh_identity(self, item, identity_field=None):
        identity = {key: None for key in ['username', 'name', 'email']}

        user = item  # by default a specific user dict is expected
        if isinstance(item, dict) and 'data' in item:
            user = item['data'][identity_field]
        elif 'author' in item:
            user = item['author']

        if user is None:
            return identity

        if 'username' in user:
            identity['username'] = user['username']
        elif 'user_display_name' in user:
            identity['username'] = user['user_display_name']
        elif 'answered_by' in user:
            if 'username' in user['answered_by']:
                identity['username'] = user['answered_by']['username']
            else:
                # "answered_by" : "This post is a wiki"
                identity['username'] = user['answered_by']
        identity['name'] = identity['username']

        return identity

    def get_field_author(self):
        return 'author'

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        question = item['data']

        if 'accepted_answer_id' not in question:
            question['accepted_answer_id'] = None

        # Fields that are the same in item and eitem
        copy_fields = ["id", "url", "title", "summary", "score"]
        for f in copy_fields:
            if f in question:
                eitem[f] = question[f]
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"title": "question_title",
                      "answer_count": "question_answer_count",
                      "view_count": "question_view_count",
                      "answer_ids": "question_answer_ids"}
        for fn in map_fields:
            if fn in question:
                eitem[map_fields[fn]] = question[fn]
            else:
                eitem[map_fields[fn]] = None

        # Cast id of question to string
        eitem['id'] = str(eitem['id'])
        eitem['score'] = int(eitem['score']) if eitem['score'] else 0

        # First answer time
        added_at = unixtime_to_datetime(float(question["added_at"]))
        eitem['time_to_reply'] = None
        if 'answers' in question:
            # answers ordered by time
            first_answer_time = unixtime_to_datetime(float(question['answers'][0]["added_at"]))
            eitem['time_to_reply'] = get_time_diff_days(added_at, first_answer_time)
            eitem['question_has_accepted_answer'] = 1 if question['accepted_answer_id'] else 0
            eitem['question_accepted_answer_id'] = question['accepted_answer_id']
        else:
            eitem['question_has_accepted_answer'] = 0

        if question['author'] and type(question['author']) is dict:
            eitem['author_askbot_user_name'] = question['author']['username']
            eitem['author_askbot_id'] = str(question['author']['id'])
            eitem['author_badges'] = question['author']['badges']
            eitem['author_reputation'] = int(question['author']['reputation'])
            eitem['author_url'] = eitem['origin'] + '/users/'
            eitem['author_url'] += question['author']['id'] + '/' + question['author']['username']

        eitem['question_last_activity_at'] = unixtime_to_datetime(float(question['last_activity_at'])).isoformat()
        eitem['question_last_activity_by_id'] = question['last_activity_by']['id']
        eitem['question_last_activity_by_username'] = question['last_activity_by']['username']
        # A list can be used directly to filter in kibana
        eitem['question_tags'] = question['tags']
        eitem['question_answer_ids'] = question['answer_ids']

        eitem['comment_count'] = 0
        if 'answers' in question:
            eitem['comment_count'] = sum([len(a['comments']) if 'comments' in a else 0 for a in question['answers']])

        if self.sortinghat:
            eitem.update(self.get_item_sh(item))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem["type"] = "question"
        eitem.update(self.get_grimoire_fields(added_at.isoformat(), eitem["type"]))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def get_field_unique_id(self):
        return "id"

    def get_users_data(self, askbot_item):
        """ Adapt the data to be used with standard SH enrich API """
        # askbot_item could be a raw question, answer or comment
        poster = {}
        poster[self.get_field_author()] = askbot_item
        return poster

    def get_rich_comment(self, item, answer, comment):
        ecomment = self.get_rich_item(item)  # reuse all fields from item
        ecomment['id'] = str(ecomment['id']) + '_' + str(answer['id']) + '_' + str(comment['id'])
        ecomment['url'] = item['data']['url'] + "/?answer="
        ecomment['url'] += answer['id'] + '#post-id-' + answer['id']
        if 'author' in comment:
            # Not sure if this format is present in some version of askbot
            ecomment['author_askbot_user_name'] = comment['author']['username']
            ecomment['author_askbot_id'] = str(comment['author']['id'])
            ecomment['author_url'] = ecomment['origin'] + '/users/'
            ecomment['author_url'] += comment['author']['id'] + '/' + comment['author']['username']

        elif 'user_display_name' in comment:
            ecomment['author_askbot_user_name'] = comment['user_display_name']
            ecomment['author_askbot_id'] = str(comment['user_id'])
        if 'summary' in comment:
            ecomment['summary'] = comment['summary']
        ecomment['score'] = int(comment['score']) if comment['score'] else 0

        dfield = 'added_at'
        if 'comment_added_at' in comment:
            dfield = 'comment_added_at'

        if self.sortinghat:
            if dfield == 'added_at':
                comment['added_at_date'] = unixtime_to_datetime(float(comment[dfield])).isoformat()
            else:
                comment['added_at_date'] = comment[dfield]
            ecomment.update(self.get_item_sh(comment, date_field="added_at_date"))
            if ecomment['author_user_name'] != ecomment['author_askbot_user_name']:
                logger.warning('[asknot] Bad SH identity in askbot comment. Found {} expecting {}'.format(
                               ecomment['author_user_name'], ecomment['author_askbot_user_name']))

        if dfield == 'added_at':
            comment_at = unixtime_to_datetime(float(comment[dfield]))
        else:
            comment_at = str_to_datetime(comment[dfield])

        added_at = unixtime_to_datetime(float(item['data']["added_at"]))
        ecomment['time_from_question'] = get_time_diff_days(added_at, comment_at)
        ecomment['type'] = 'comment'
        ecomment.update(self.get_grimoire_fields(comment_at.isoformat(), ecomment['type']))

        # Clean items fields not valid in comments
        for f in ['is_askbot_question', 'author_reputation', 'author_badges', 'is_correct', 'comment_count']:
            if f in ecomment:
                ecomment.pop(f)

        return ecomment

    def get_rich_answer(self, item, answer):
        eanswer = self.get_rich_item(item)  # reuse all fields from item
        eanswer['id'] = str(eanswer['id']) + '_' + str(answer['id'])
        eanswer['url'] = item['data']['url'] + "/?answer="
        eanswer['url'] += answer['id'] + '#post-id-' + answer['id']
        if type(answer['answered_by']) is dict:
            eanswer['author_askbot_user_name'] = answer['answered_by']['username']
            eanswer['author_askbot_id'] = str(answer['answered_by']['id'])
            eanswer['author_badges'] = answer['answered_by']['badges']
            eanswer['author_reputation'] = int(answer['answered_by']['reputation'])
            eanswer['author_url'] = eanswer['origin'] + '/users/'
            eanswer['author_url'] += answer['answered_by']['id'] + '/'
            eanswer['author_url'] += answer['answered_by']['username']

        eanswer['summary'] = answer['summary']
        eanswer['is_accepted_answer'] = 1 if answer['accepted'] else 0
        eanswer['answer_status'] = "accepted" if answer['accepted'] else "not_accepted"
        eanswer['score'] = int(answer['score']) if answer['score'] else 0
        if 'is_correct' in answer:
            eanswer['is_correct'] = 1

        if self.sortinghat:
            answer['added_at_date'] = unixtime_to_datetime(float(answer["added_at"])).isoformat()
            eanswer.update(self.get_item_sh(answer, date_field="added_at_date"))
            if 'author_askbot_user_name' in eanswer and eanswer['author_user_name'] != eanswer['author_askbot_user_name']:
                logger.warning('[askbot] Bad SH identity in askbot answer. Found {} expecting {}'.format(
                               eanswer['author_user_name'], eanswer['author_askbot_user_name']))
        answer_at = unixtime_to_datetime(float(answer["added_at"]))
        added_at = unixtime_to_datetime(float(item['data']["added_at"]))
        eanswer['time_from_question'] = get_time_diff_days(added_at, answer_at)
        eanswer['type'] = 'answer'
        eanswer.update(self.get_grimoire_fields(answer_at.isoformat(), eanswer['type']))

        # Clean items fields not valid in comments
        eanswer.pop('is_askbot_question')

        return eanswer

    def get_rich_item_answers_comments(self, item):
        answers_enrich = []
        comments_enrich = []

        if 'answers' not in item['data']:
            return answers_enrich, comments_enrich

        for answer in item['data']['answers']:
            eanswer = self.get_rich_answer(item, answer)
            if not answers_enrich:
                eanswer['first_answer'] = 1
            answers_enrich.append(eanswer)

            # And now time to process the comments
            if 'comments' in answer:
                for comment in answer['comments']:
                    ecomment = self.get_rich_comment(item, answer, comment)
                    comments_enrich.append(ecomment)

        return answers_enrich, comments_enrich

    def enrich_items(self, ocean_backend):
        items_to_enrich = []
        num_items = 0
        ins_items = 0

        for item in ocean_backend.fetch():
            eitem = self.get_rich_item(item)
            items_to_enrich.append(eitem)

            (answers, comments) = self.get_rich_item_answers_comments(item)
            items_to_enrich.extend(answers)
            items_to_enrich.extend(comments)

            if len(items_to_enrich) < MAX_SIZE_BULK_ENRICHED_ITEMS:
                continue

            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())
            items_to_enrich = []

        if len(items_to_enrich) > 0:
            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())

        if num_items != ins_items:
            missing = num_items - ins_items
            logger.error("[askbot] {}/{} missing items for Askbot".format(missing, num_items))
        else:
            logger.info("[askbot] {} items inserted for Askbot".format(num_items))

        return num_items



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Valerio Cosentino <valcos@bitergia.com>
#   Miguel Ángel Fernández <mafesan@bitergia.com>
#   Quan Zhou <quan@bitergia.com>
#

import logging
import re

from elasticsearch import Elasticsearch as ES, RequestsHttpConnection

from .enrich import Enrich, metadata
from .utils import anonymize_url, get_time_diff_days
from ..elastic_mapping import Mapping as BaseMapping

GITHUB = 'https://github.com/'
LABEL_EVENTS = ['LabeledEvent', 'UnlabeledEvent']
PROJECT_EVENTS = ['AddedToProjectEvent', 'MovedColumnsInProjectEvent', 'RemovedFromProjectEvent']
REFERENCE_EVENTS = ['CrossReferencedEvent']
CLOSED_EVENTS = ['ClosedEvent']
MERGED_EVENTS = ['MergedEvent']
PULL_REQUEST_REVIEW_EVENTS = ['PullRequestReview']

logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        geopoints type is not created in dynamic mapping

        :param es_major: major version of Elasticsearch, as string
        :returns:        dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
               "issue_state": {
                   "type": "keyword"
               },
               "title_analyzed": {
                 "type": "text",
                 "index": true
               }
            }
        }
        """

        return {"items": mapping}


class GitHubQLEnrich(Enrich):

    mapping = Mapping

    event_roles = ['actor', 'reporter', 'submitter']

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                 db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map,
                         db_user, db_password, db_host)

        self.studies = [self.enrich_duration_analysis, self.enrich_reference_analysis]

    def set_elastic(self, elastic):
        self.elastic = elastic

    def get_field_author(self):
        return "actor"

    def get_field_date(self):
        """ Field with the date in the JSON enriched items """
        return "grimoire_creation_date"

    def get_identities(self, item):
        """Return the identities from an item"""

        event = item['data']
        event_actor = event.get("actor", event.get("author", None))
        if event_actor:
            identity = self.get_sh_identity(event_actor)
            if identity:
                yield identity

        issue = event['issue']
        issue_reporter = issue.get("user", None)
        if issue_reporter:
            identity = self.get_sh_identity(issue_reporter)
            if identity:
                yield identity

        closer = event.get('closer', None)
        if closer and closer['type'] == 'PullRequest':
            pull_submitter = closer.get('author', None)
            identity = self.get_sh_identity(pull_submitter)
            if identity:
                yield identity

    def get_sh_identity(self, item, identity_field=None):
        identity = {}

        user = item  # by default a specific user dict is expected
        if isinstance(item, dict) and 'data' in item:
            if identity_field == 'actor':
                user = item['data'][identity_field]
            elif identity_field == 'reporter':
                user = item['data']['issue']['user']
            elif identity_field == 'submitter':
                closer = item['data'].get('closer', None)
                if closer:
                    user = closer['author']

        if not user:
            return identity

        identity['username'] = user['login']
        identity['email'] = None
        identity['name'] = None

        return identity

    def get_project_repository(self, eitem):
        repo = eitem['origin']
        return repo

    @metadata
    def get_rich_item(self, item):

        rich_item = self.__get_rich_event(item)

        self.add_repository_labels(rich_item)
        self.add_metadata_filter_raw(rich_item)

        return rich_item

    def __get_rich_event(self, item):
        rich_event = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, rich_event)

        event = item['data']
        issue = item['data']['issue']
        actor = item['data'].get('actor', item['data'].get('author', None))

        # move the issue reporter to level of actor. This is needed to
        # allow `get_item_sh` adding SortingHat identities
        reporter = issue['user']
        item['data']['reporter'] = reporter

        rich_event['event_type'] = event['eventType']
        rich_event['created_at'] = event['createdAt']
        rich_event['actor_username'] = actor['login'] if actor else None
        rich_event['repository'] = self.get_project_repository(rich_event)
        rich_event['pull_request'] = True
        rich_event['item_type'] = 'pull request'
        if 'head' not in issue.keys() and 'pull_request' not in issue.keys():
            rich_event['pull_request'] = False
            rich_event['item_type'] = 'issue'

        rich_event['issue_id'] = issue['id']
        rich_event['issue_id_in_repo'] = issue['html_url'].split("/")[-1]
        rich_event['title'] = issue['title']
        rich_event['title_analyzed'] = issue['title']
        rich_event['issue_state'] = issue['state']
        rich_event['issue_created_at'] = issue['created_at']
        rich_event['issue_updated_at'] = issue['updated_at']
        rich_event['issue_closed_at'] = issue['closed_at']
        rich_event['issue_url'] = issue['html_url']
        labels = []
        [labels.append(label['name']) for label in issue['labels'] if 'labels' in issue]
        rich_event['issue_labels'] = labels

        rich_event['github_repo'] = rich_event['repository'].replace(GITHUB, '')
        rich_event['github_repo'] = re.sub('.git$', '', rich_event['github_repo'])
        rich_event["issue_url_id"] = rich_event['github_repo'] + "/issues/" + rich_event['issue_id_in_repo']

        if rich_event['event_type'] in LABEL_EVENTS:
            label = event['label']
            rich_event['label'] = label['name']
            rich_event['label_description'] = label['description']
            rich_event['label_is_default'] = label['isDefault']
            rich_event['label_created_at'] = label['createdAt']
            rich_event['label_updated_at'] = label['updatedAt']
        elif rich_event['event_type'] in CLOSED_EVENTS:
            closer = event['closer']
            rich_event['label'] = rich_event['issue_labels']
            # In GitHub every pull request is an issue. When retrieving closed events for
            # issues using the GraphQL API, the closed events related to pull requests are
            # collected too. Since the attribute closer is not defined for pull requests,
            # the condition below makes sure to prevent NPE errors.
            # The test data at tests/data/githubql.json contains two examples with the
            # attribute closer being null and not null.
            if closer and closer['type'] == 'PullRequest':
                rich_event['closer_event_url'] = event['url']
                rich_event['closer_type'] = closer['type']
                rich_event['closer_number'] = closer['number']
                rich_event['closer_url'] = closer['url']
                rich_event['closer_repo'] = '/'.join(closer['url'].replace(GITHUB, '').split('/')[:-2])
                rich_event['closer_created_at'] = closer['createdAt']
                rich_event['closer_updated_at'] = closer['updatedAt']
                rich_event['closer_closed_at'] = closer['closedAt']
                rich_event['closer_closed'] = closer['closed']
                rich_event['closer_merged'] = closer.get('merged', None)
                submitter = closer['author']
                rich_event['closer_pull_submitter'] = submitter.get('login', None) if submitter else None
                # move the pull request submitter to level of actor. This is needed to
                # allow `get_item_sh` adding SortingHat identities
                item['data']['submitter'] = submitter
        elif rich_event['event_type'] in REFERENCE_EVENTS:
            source = event['source']
            rich_event['reference_cross_repo'] = event['isCrossRepository']
            rich_event['reference_will_close_target'] = event['willCloseTarget']
            rich_event['reference_event_url'] = event['url']
            rich_event['reference_source_type'] = source['type']
            rich_event['reference_source_number'] = source['number']
            rich_event['reference_source_url'] = source['url']
            rich_event['reference_source_repo'] = '/'.join(source['url'].replace(GITHUB, '').split('/')[:-2])
            rich_event['reference_source_created_at'] = source['createdAt']
            rich_event['reference_source_updated_at'] = source['updatedAt']
            rich_event['reference_source_closed_at'] = source['closedAt']
            rich_event['reference_source_closed'] = source['closed']
            rich_event['reference_source_merged'] = source.get('merged', None)
        elif rich_event['event_type'] in PROJECT_EVENTS:
            project = event['project']
            rich_event['board_column'] = event['projectColumnName']
            rich_event['board_name'] = project['name']
            rich_event['board_url'] = project['url']
            rich_event['board_created_at'] = project['createdAt']
            rich_event['board_updated_at'] = project['updatedAt']
            rich_event['board_closed_at'] = project['closedAt']
            rich_event['board_state'] = project['state'].lower()

            # only for events of type MovedColumnsInProjectEvent
            if 'previousProjectColumnName' in event:
                rich_event['board_previous_column'] = event['previousProjectColumnName']
        elif rich_event['event_type'] in MERGED_EVENTS:
            merge = event['pullRequest']
            rich_event['merge_closed'] = merge['closed']
            rich_event['merge_closed_at'] = merge['closedAt']
            rich_event['merge_created_at'] = merge['createdAt']
            rich_event['merge_merged'] = merge['merged']
            rich_event['merge_merged_at'] = merge['mergedAt']
            rich_event['merge_updated_at'] = merge['updatedAt']
            rich_event['merge_url'] = merge['url']
        elif rich_event['event_type'] in PULL_REQUEST_REVIEW_EVENTS:
            review = event['pullRequest']
            rich_event['merge_approved'] = 0
            rich_event['merge_state'] = event['state']
            rich_event['merge_approved'] = int(rich_event['merge_state'] == 'APPROVED')
            rich_event['merge_closed'] = review['closed']
            rich_event['merge_closed_at'] = review['closedAt']
            rich_event['merge_created_at'] = review['createdAt']
            rich_event['merge_merged'] = review['merged']
            rich_event['merge_merged_at'] = review['mergedAt']
            rich_event['merge_updated_at'] = review['updatedAt']
            rich_event['merge_url'] = review['url']
            item['data']['actor'] = item['data']['author']
        else:
            logger.warning("[github] event {} not processed".format(rich_event['event_type']))

        if self.prjs_map:
            rich_event.update(self.get_item_project(rich_event))

        rich_event.update(self.get_grimoire_fields(event['createdAt'], "issue"))
        item[self.get_field_date()] = rich_event[self.get_field_date()]
        rich_event.update(self.get_item_sh(item, self.event_roles))

        # Copy SH actor info to author equivalent attributes
        rich_event['author_id'] = rich_event.get('actor_id', None)
        rich_event['author_uuid'] = rich_event.get('actor_uuid', None)
        rich_event['author_name'] = rich_event.get('actor_name', None)
        rich_event['author_user_name'] = rich_event.get('actor_user_name', None)
        rich_event['author_domain'] = rich_event.get('actor_domain', None)
        rich_event['author_gender'] = rich_event.get('actor_gender', None)
        rich_event['author_gender_acc'] = rich_event.get('actor_gender_acc', None)
        rich_event['author_org_name'] = rich_event.get('actor_org_name', None)
        rich_event['author_bot'] = rich_event.get('actor_bot', None)
        rich_event['author_multi_org_names'] = rich_event.get('actor_multi_org_names', None)

        return rich_event

    def enrich_duration_analysis(self, ocean_backend, enrich_backend, start_event_type, target_attr,
                                 fltr_event_types, fltr_attr=None, page_size=200):
        """The purpose of this study is to calculate the duration between two GitHub events. It requires
        a start event type (e.g., UnlabeledEvent or MovedColumnsInProjectEvent), which is used to
        retrieve for each issue all events of that type. For each issue event obtained, the first
        previous event of one of the types defined at `fltr_event_types` is returned, and used to
        calculate the duration (in days) between the two events. Optionally, an additional filter
        can be defined to retain the events that share a given property (e.g., a specific label,
        the name of project board). Finally, the duration and the previous event uuid are added to
        the start event via the attributes `duration_from_previous_event` and `previous_event_uuid`.

        This study is executed in a incremental way, thus only the start events that don't
        include the attribute `duration_from_previous_event` are retrieved and processed.

        The examples below show how to activate the study by modifying the setup.cfg. The first example
        calculates the duration between Unlabeled and Labeled events per label. The second example
        calculates the duration between the MovedColumnsInProject and AddedToProject events per
        column in each board

        ```
        [githubql]
        ...
        studies = [enrich_duration_analysis:label, enrich_duration_analysis:kanban]

        [enrich_duration_analysis:kanban]
        start_event_type = MovedColumnsInProjectEvent
        fltr_attr = board_name
        target_attr = board_column
        fltr_event_types = [MovedColumnsInProjectEvent, AddedToProjectEvent]

        [enrich_duration_analysis:label]
        start_event_type = UnlabeledEvent
        target_attr = label
        fltr_attr = label
        fltr_event_types = [LabeledEvent]
        ```

        :param ocean_backend: backend from which to read the raw items
        :param enrich_backend:  backend from which to read the enriched items
        :param start_event_type: the type of the start event (e.g., UnlabeledEvent)
        :param target_attr: the attribute returned from the events (e.g., label)
        :param fltr_event_types: a list of event types to select the previous events (e.g., LabeledEvent)
        :param fltr_attr: an optional attribute to filter in the events with a given property (e.g., label)
        :param page_size: number of events without `duration_from_previous_event` per page
        """
        data_source = enrich_backend.__class__.__name__.split("Enrich")[0].lower()
        log_prefix = "[{}] Duration analysis".format(data_source)
        logger.info("{} starting study {}".format(log_prefix, anonymize_url(self.elastic.index_url)))

        es_in = ES([enrich_backend.elastic_url], retry_on_timeout=True, timeout=100,
                   verify_certs=self.elastic.requests.verify, connection_class=RequestsHttpConnection)
        in_index = enrich_backend.elastic.index

        # get all start events that don't have the attribute `duration_from_previous_event`
        query_start_event_type = {
            "query": {
                "bool": {
                    "filter": {
                        "term": {
                            "event_type": start_event_type
                        }
                    },
                    "must_not": {
                        "exists": {
                            "field": "duration_from_previous_event"
                        }
                    }
                }
            },
            "_source": [
                "uuid", "issue_url_id", "grimoire_creation_date", target_attr
            ],
            "sort": [
                {
                    "grimoire_creation_date": {
                        "order": "asc"
                    }
                }
            ],
            "size": page_size
        }

        if fltr_attr:
            query_start_event_type['_source'].append(fltr_attr)

        start_event_types = es_in.search(index=in_index, body=query_start_event_type, scroll='5m')

        sid = start_event_types['_scroll_id']
        scroll_size = len(start_event_types['hits']['hits'])

        while scroll_size > 0:

            # for each event, retrieve the previous event included in `fltr_event_types`
            for start_event in start_event_types['hits']['hits']:
                start_event = start_event['_source']
                start_uuid = start_event['uuid']
                start_issue_url_id = start_event['issue_url_id']
                start_date_event = start_event['grimoire_creation_date']

                query_previous_events = {
                    "size": 1,
                    "query": {
                        "bool": {
                            "filter": [
                                {
                                    "term": {
                                        "issue_url_id": start_issue_url_id
                                    }
                                },
                                {
                                    "terms": {
                                        "event_type": fltr_event_types
                                    }
                                },
                                {
                                    "range": {
                                        "grimoire_creation_date": {
                                            "lt": start_date_event
                                        }
                                    }
                                }
                            ]
                        }
                    },
                    "_source": [
                        "uuid", "grimoire_creation_date", target_attr
                    ],
                    "sort": [
                        {
                            "grimoire_creation_date": {
                                "order": "desc"
                            }
                        }
                    ]
                }

                if fltr_attr:
                    _fltr = {
                        "term": {
                            fltr_attr: start_event[fltr_attr]
                        }
                    }

                    query_previous_events['query']['bool']['filter'].append(_fltr)
                    query_start_event_type['_source'].append(fltr_attr)

                previous_events = es_in.search(index=in_index, body=query_previous_events)['hits']['hits']
                if not previous_events:
                    continue

                previous_event = previous_events[0]['_source']
                previous_event_date = previous_event['grimoire_creation_date']
                previous_event_uuid = previous_event['uuid']
                duration = get_time_diff_days(previous_event_date, start_date_event)

                painless_code = "ctx._source.duration_from_previous_event=params.duration;" \
                                "ctx._source.previous_event_uuid=params.uuid"

                add_previous_event_query = {
                    "script": {
                        "source": painless_code,
                        "lang": "painless",
                        "params": {
                            "duration": duration,
                            "uuid": previous_event_uuid
                        }
                    },
                    "query": {
                        "bool": {
                            "filter": {
                                "term": {
                                    "uuid": start_uuid
                                }
                            }
                        }
                    }
                }
                r = es_in.update_by_query(index=in_index, body=add_previous_event_query, conflicts='proceed')
                if r['failures']:
                    logger.error("{} Error while executing study {}".format(log_prefix,
                                                                            anonymize_url(self.elastic.index_url)))
                    logger.error(str(r['failures'][0]))
                    return

            start_event_types = es_in.scroll(scroll_id=sid, scroll='2m')
            # update the scroll ID
            sid = start_event_types['_scroll_id']
            # get the number of results that returned in the last scroll
            scroll_size = len(start_event_types['hits']['hits'])

        logger.info("{} ending study {}".format(log_prefix, anonymize_url(self.elastic.index_url)))

    def enrich_reference_analysis(self, ocean_backend, enrich_backend, aliases_update=None):
        """
        The purpose of this study is to gather all the issues and pull requests which are
        mutually referenced. Once these references are obtained, all of the events for the given issue
        or pull request are updated with the corresponding list of URLs from the referenced items.

        This study is not executed in a incremental way, as it only takes the `CrossReferencedEvent`
        items to build a dictionary containing all the mutual references per each Issue or Pull Request,
        identified by `issue_url`. Then, it updates all the events belonging to the same `issue_url`
        adding the following fields:
        * `referenced_by_issues`: List of issues referenced by a given Issue or Pull Request,
            from the same repository.
        * `referenced_by_prs`: List of pull requests referenced by a given Issue or Pull Request,
            from the same repository.
          `referenced_by_merged_prs`: List of merged pull requests referenced by a given Issue or Pull
            Request, from the same repository.
        * `referenced_by_external_issues`: List of issues referenced by a given Issue or Pull Request,
            from different (external) repositories.
        * `referenced_by_external_prs`: List of pull requests referenced by a given Issue or Pull Request,
            from different (external) repositories.
        * `referenced_by_external_merged_prs`: List of merged pull requests referenced by a given Issue
            or Pull Request, from different (external) repositories.

        To classify the merged Pull Requests, the study asks for the list of the URLs from `MergedEvents`.

        The method accepts a list of ES aliases or indices where these new fields will be updated too,
        for the referenced elements identified by a given `issue_url`. An example would be the `github_issues`
        alias, which points to the corresponding enriched indexes from GitHub issues. Note that if the
        referenced aliases or indexes do not contain the `issue_url` field, the affected elements can't
        be updated.

        The examples below show how to activate the study by modifying the setup.cfg.

        ```
        [githubql]
        ...
        studies = [..., enrich_reference_analysis]

        [enrich_reference_analysis]]
        aliases_update = [github_issues, github2_issues, github2_pull_requests]
        ```

        :param ocean_backend: backend from which to read the raw items
        :param enrich_backend:  backend from which to read the enriched items
        :param aliases_update: list of aliases where to update the referenced items
        """
        def _is_pull_request(issue_url):
            """Return True if `issue_url` belongs to a Pull Request"""

            return '/pull/' in issue_url

        def _get_github_repo(issue_url):
            """Return GitHub repository path using `owner/repository` format"""

            repo = ''
            url_info = issue_url.split('/')
            if url_info[2] == 'github.com':
                repo = '/'.join([url_info[3], url_info[4]])
            return repo

        def _get_merged_prs(es_input):
            """Return a list of merged Pull Requests based on MergedEvent items"""

            # Ask for the URL from `MergedEvent` items, filtering by merged PRs
            es_query = {
                "size": 0,
                "query": {
                    "bool": {
                        "must": [
                            {
                                "term": {"event_type": "MergedEvent"}
                            },
                            {
                                "term": {"pull_request": True}
                            },
                            {
                                "term": {"merge_merged": True}
                            }
                        ]
                    }
                },
                "aggs": {
                    "merge_url": {
                        "terms": {
                            "field": "merge_url",
                            "size": 30000
                        }
                    }
                }
            }

            merged_prs = es_input.search(index=in_index, body=es_query)
            buckets = merged_prs['aggregations']['merge_url']['buckets']

            merged_prs_list = [item['key'] for item in buckets]

            return merged_prs_list

        data_source = enrich_backend.__class__.__name__.split("Enrich")[0].lower()
        log_prefix = "[{}] Cross reference analysis".format(data_source)
        logger.info("{} starting study {}".format(log_prefix, anonymize_url(self.elastic.index_url)))

        es_in = ES([enrich_backend.elastic_url], retry_on_timeout=True, timeout=100,
                   verify_certs=self.elastic.requests.verify, connection_class=RequestsHttpConnection)
        in_index = enrich_backend.elastic.index

        # Get all the merged pull requests from MergedEvents
        logger.info("{} Retrieving the merged PRs from MergeEvents".format(log_prefix))
        merged_prs = _get_merged_prs(es_in)

        # Get all CrossReferencedEvent items and their referenced issues and pull requests
        es_query = {
            "size": 0,
            "query": {
                "bool": {
                    "must": {
                        "term": {
                            "event_type": "CrossReferencedEvent"
                        }
                    }
                }
            },
            "aggs": {
                "issue_url": {
                    "terms": {
                        "field": "issue_url",
                        "size": 30000
                    },
                    "aggs": {
                        "uniq_gender": {
                            "terms": {"field": "reference_source_url"}
                        }
                    }
                }
            }
        }

        cross_references = es_in.search(index=in_index, body=es_query)
        buckets = cross_references['aggregations']['issue_url']['buckets']

        reference_dict = {}
        for item in buckets:
            issue_url = item['key']
            references = [ref['key'] for ref in item['uniq_gender']['buckets']]

            # Update reference dictionary
            if issue_url not in reference_dict.keys():
                reference_dict[issue_url] = references
            else:
                prev_references = reference_dict[issue_url]
                prev_references.append(references)
                reference_dict[issue_url] = list(set(prev_references))

        # Adding list entries from reversed references
        for issue_url in reference_dict.keys():
            reference_list = reference_dict[issue_url]
            if not reference_list:
                continue
            for ref in reference_list:
                try:
                    ref_entry_list = reference_dict[ref]
                except KeyError:
                    continue
                if ref_entry_list:
                    ref_entry_list.append(issue_url)
                else:
                    ref_entry_list = [issue_url]
                reference_dict[ref] = list(set(ref_entry_list))

        # Updated affected issues and pull requests
        painless_code = """
            ctx._source.referenced_by_issues = params.referenced_by_issues;
            ctx._source.referenced_by_prs = params.referenced_by_prs;
            ctx._source.referenced_by_merged_prs = params.referenced_by_merged_prs;
            ctx._source.referenced_by_external_issues = params.referenced_by_external_issues;
            ctx._source.referenced_by_external_prs = params.referenced_by_external_prs;
            ctx._source.referenced_by_external_merged_prs = params.referenced_by_external_merged_prs;
        """
        for issue_url in reference_dict.keys():
            ref_issues_repo = []
            ref_prs_repo = []
            ref_prs_merged_repo = []
            ref_issues_ext = []
            ref_prs_ext = []
            ref_prs_merged_ext = []

            issue_repo = _get_github_repo(issue_url)

            # Classify references internal/external repo + issues/pull-requests
            reference_list = reference_dict[issue_url]
            for ref in reference_list:
                ref_repo = _get_github_repo(ref)
                ref_is_pr = _is_pull_request(ref)

                # Classify references
                if ref_repo == issue_repo:
                    if ref_is_pr:
                        ref_prs_repo.append(ref)
                        if ref in merged_prs:
                            ref_prs_merged_repo.append(ref)
                    else:
                        ref_issues_repo.append(ref)
                else:
                    if ref_is_pr:
                        ref_prs_ext.append(ref)
                        if ref in merged_prs:
                            ref_prs_merged_ext.append(ref)
                    else:
                        ref_issues_ext.append(ref)

            # Update items with the corresponding fields
            update_query = {
                "script": {
                    "source": painless_code,
                    "lang": "painless",
                    "params": {
                        "referenced_by_issues": ref_issues_repo,
                        "referenced_by_prs": ref_prs_repo,
                        "referenced_by_merged_prs": ref_prs_merged_repo,
                        "referenced_by_external_issues": ref_issues_ext,
                        "referenced_by_external_prs": ref_prs_ext,
                        "referenced_by_external_merged_prs": ref_prs_merged_ext,
                    }
                },
                "query": {
                    "term": {
                        "issue_url": issue_url
                    }
                }
            }

            update_indexes = [in_index]
            # Update data in the additional related indexes (if any) identified by their aliases
            if aliases_update:
                update_indexes += aliases_update

            for update_index in update_indexes:
                logger.info('{} - Updating fields from items with issue_url: {} from index {}'.format(log_prefix,
                                                                                                      issue_url,
                                                                                                      update_index))

                r = es_in.update_by_query(index=update_index, body=update_query, conflicts='proceed')
                if r['failures']:
                    logger.error("{} Error while executing study {}".format(log_prefix,
                                                                            anonymize_url(self.elastic.index_url)))
                    logger.error(str(r['failures'][0]))
                    return

        logger.info("{} ending study {}".format(log_prefix, anonymize_url(self.elastic.index_url)))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging

from ..elastic_mapping import Mapping as BaseMapping
from .utils import get_time_diff_days, grimoire_con

from .enrich import Enrich, metadata

MAX_SIZE_BULK_ENRICHED_ITEMS = 200
# Headers
HKEY = 'Api-Key'
HUSER = 'Api-Username'


logger = logging.getLogger(__name__)


class Mapping(BaseMapping):

    @staticmethod
    def get_elastic_mappings(es_major):
        """Get Elasticsearch mapping.

        :param es_major: major version of Elasticsearch, as string
        :returns: dictionary with a key, 'items', with the mapping
        """

        mapping = """
        {
            "properties": {
               "question_title_analyzed": {
                    "type": "text",
                    "index": true
               }
           }
        }"""

        return {"items": mapping}


class DiscourseEnrich(Enrich):

    mapping = Mapping

    def __init__(self, db_sortinghat=None, db_projects_map=None, json_projects_map=None,
                 db_user='', db_password='', db_host=''):
        super().__init__(db_sortinghat, db_projects_map, json_projects_map,
                         db_user, db_password, db_host)
        self.categories = {}  # Map from category_id to category_name
        self.categories_tree = {}  # Categories with subcategories

    def get_identities(self, item):
        """ Return the identities from an item """

        # All identities are in the post stream
        # The first post is the question. Next replies

        posts = item['data']['post_stream']['posts']

        for post in posts:
            user = self.get_sh_identity(post)
            yield user

    def get_sh_identity(self, post, identity_field=None):
        identity = {}

        if identity_field in post:
            post = post[identity_field]

        identity['username'] = post.get('username', None)
        identity['email'] = None
        identity['name'] = post.get('display_username', None)
        return identity

    def get_field_author(self):
        return 'author'

    def get_project_repository(self, eitem):
        return str(eitem['category_id'])

    def get_users_data(self, post):
        """ Adapt the data to be used with standard SH enrich API """
        poster = {}
        poster[self.get_field_author()] = post
        return poster

    def get_rich_item_answers(self, item):
        answers_enrich = []
        nanswers = 0

        for answer in item['data']['post_stream']['posts']:
            eanswer = self.get_rich_item(item)  # reuse all fields from item
            eanswer['id'] = str(eanswer['id']) + '_' + str(answer['id'])
            eanswer['url'] = eanswer['origin'] + "/t/" + answer['topic_slug']
            eanswer['url'] += "/" + str(answer['topic_id']) + "/" + str(answer['post_number'])
            eanswer['type'] = 'answer'
            eanswer.update(self.get_grimoire_fields(answer['created_at'], eanswer['type']))
            eanswer.pop('is_discourse_question')
            eanswer['display_username'] = answer.get('display_username', None)
            eanswer['username'] = answer.get('username', None)
            eanswer['author_id'] = answer['user_id']
            eanswer['author_trust_level'] = answer['trust_level']
            eanswer['author_url'] = eanswer['origin'] + "/users/" + str(eanswer['author_id'])
            eanswer['reads'] = answer['reads']
            eanswer['score'] = float(answer['score'])
            eanswer['reply_count'] = answer['reply_count']
            eanswer['time_from_question'] = None
            post_time = answer['created_at']
            item_time = item['data']['created_at']
            eanswer['first_reply_time'] = get_time_diff_days(item_time, post_time)
            answers_enrich.append(eanswer)

            if self.sortinghat:
                eanswer.update(self.get_item_sh(answer, date_field="created_at"))

            if self.prjs_map:
                eanswer.update(self.get_item_project(eanswer))

            nanswers += 1
            eanswer['first_answer'] = 0
            if nanswers == 1:
                eanswer['first_answer'] = 1

            if 'accepted_answer' not in answer:
                answer['accepted_answer'] = False

            eanswer['is_accepted_answer'] = 1 if answer['accepted_answer'] else 0
            eanswer['answer_status'] = "accepted" if answer['accepted_answer'] else "not_accepted"

        return answers_enrich

    def _set_extra_headers(self):
        """Set extra headers for session"""
        headers = {}
        api_token = self.perceval_backend.api_token
        api_username = self.perceval_backend.api_username
        if api_token and api_username:
            headers[HKEY] = api_token
            headers[HUSER] = api_username

        return headers

    def __collect_categories(self, origin, headers):
        categories = {}
        con = grimoire_con()

        raw_site = con.get(origin + "/site.json", headers=headers)
        for cat in raw_site.json()['categories']:
            categories[cat['id']] = cat['name']
        return categories

    def __collect_categories_tree(self, origin, headers):
        tree = {}
        con = grimoire_con()
        raw = con.get(origin + "/categories.json", headers=headers)
        raw_json = raw.json()
        if "category_list" in raw_json and 'categories' in raw_json["category_list"]:
            categories = raw_json["category_list"]['categories']
        else:
            return tree
        for cat in categories:
            if 'subcategory_ids' in cat:
                tree[cat['id']] = cat['subcategory_ids']
            else:
                tree[cat['id']] = {}
        return tree

    def __related_categories(self, category_id):
        """ Get all related categories to a given one """
        related = []
        for cat in self.categories_tree:
            if category_id in self.categories_tree[cat]:
                related.append(self.categories[cat])
        return related

    @metadata
    def get_rich_item(self, item):

        # Get the categories name if not already done
        if not self.categories:
            logger.info("[discourse] Getting the categories data from {}".format(
                        item['origin']))
            self.categories = self.__collect_categories(item['origin'], self._set_extra_headers())
        # Get the categories tree if not already done
        if not self.categories_tree:
            logger.info("[discourse] Getting the categories tree data from {}".format(
                        item['origin']))
            self.categories_tree = self.__collect_categories_tree(item['origin'], self._set_extra_headers())
            # self.__show_categories_tree()

        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        topic = item['data']

        # Fields that are the same in item and eitem
        copy_fields = ["id"]
        for f in copy_fields:
            if f in topic:
                eitem[f] = str(topic[f])
            else:
                eitem[f] = None
        # Fields which names are translated
        map_fields = {"like_count": "question_like_count",
                      "posts_count": "question_posts_count",
                      "participant_count": "question_participants",
                      "pinned_at": "question_pinned_at",
                      "pinned_globally": "question_pinned_globally",
                      "pinned_until": "question_pinned_until",
                      "pinned": "question_pinned",
                      "fancy_title": "question_title",
                      "views": "question_views",
                      "reply_count": "question_replies"
                      }
        for fn in map_fields:
            if fn in topic:
                eitem[map_fields[fn]] = topic[fn]
            else:
                eitem[map_fields[fn]] = None

        # The first post is the first published, and it is the question
        first_post = topic['post_stream']['posts'][0]

        eitem['question_title'] = eitem['question_title'][:self.KEYWORD_MAX_LENGTH]
        eitem['question_title_analyzed'] = eitem['question_title']

        eitem['category_id'] = topic['category_id']
        eitem['categories'] = self.__related_categories(topic['category_id'])
        if topic['category_id'] in self.categories:
            eitem['category_name'] = self.categories[topic['category_id']]
            eitem['categories'] += [eitem['category_name']]
        eitem['url'] = eitem['origin'] + "/t/" + first_post['topic_slug']
        eitem['url'] += "/" + str(first_post['topic_id']) + "/" + str(first_post['post_number'])
        eitem['display_username'] = first_post.get('display_username', None)
        eitem['username'] = first_post.get('username', None)
        eitem['author_id'] = first_post['user_id']
        eitem['author_trust_level'] = first_post['trust_level']
        eitem['author_url'] = eitem['origin'] + "/users/" + str(eitem['author_id'])
        eitem['reads'] = first_post['reads']
        eitem['score'] = float(first_post['score'])
        eitem['reply_count'] = first_post['reply_count']

        eitem['question_has_accepted_answer'] = 0
        eitem['question_accepted_answer_id'] = None
        # First reply time
        eitem['time_from_question'] = None
        firt_post_time = None
        if len(topic['post_stream']['posts']) > 1:
            firt_post_time = first_post['created_at']
            second_post_time = topic['post_stream']['posts'][1]['created_at']
            eitem['first_reply_time'] = get_time_diff_days(firt_post_time, second_post_time)
            answers_id = [p['id'] for p in topic['post_stream']['posts']
                          if 'accepted_answer' in p and p['accepted_answer']]
            eitem['question_accepted_answer_id'] = answers_id[0] if answers_id else None
            eitem['question_has_accepted_answer'] = 1 if eitem['question_accepted_answer_id'] else 0

        if self.sortinghat:
            eitem.update(self.get_item_sh(first_post, date_field="created_at"))

        if self.prjs_map:
            eitem.update(self.get_item_project(eitem))

        eitem['type'] = 'question'
        eitem.update(self.get_grimoire_fields(topic["created_at"], eitem['type']))

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    def get_field_unique_id(self):
        return "id"

    def enrich_items(self, ocean_backend):
        items_to_enrich = []
        num_items = 0
        ins_items = 0

        for item in ocean_backend.fetch():
            eitem = self.get_rich_item(item)
            items_to_enrich.append(eitem)

            rich_item_answers = self.get_rich_item_answers(item)
            items_to_enrich.extend(rich_item_answers)

            if len(items_to_enrich) < MAX_SIZE_BULK_ENRICHED_ITEMS:
                continue

            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())
            items_to_enrich = []

        if len(items_to_enrich) > 0:
            num_items += len(items_to_enrich)
            ins_items += self.elastic.bulk_upload(items_to_enrich, self.get_field_unique_id())

        if num_items != ins_items:
            missing = num_items - ins_items
            logger.error("[discourse] {}/{} missing items for Discourse".format(
                         missing, num_items))
        else:
            logger.info("[discourse] {} items inserted for Discourse".format(num_items))

        return num_items



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2019 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Alvaro del Castillo San Felix <acs@bitergia.com>
#

import logging
import re

from grimoirelab_toolkit.datetime import str_to_datetime, unixtime_to_datetime

from .enrich import Enrich, metadata


logger = logging.getLogger(__name__)


class FunctestEnrich(Enrich):

    BOOST_PROJECTS = ['functest', 'storperf', 'vsperf', 'bottlenecks', 'qtip', 'yardstick']

    def get_identities(self, item):
        """ Return the identities from an item """
        identities = []

        return identities

    def has_identities(self):
        """ Return whether the enriched items contains identities """

        return False

    def get_field_author(self):
        # In Functest there is no identities support
        return None

    def __process_duration(self, start_date, stop_date):
        difference_in_seconds = None

        if not start_date or not stop_date:
            return difference_in_seconds

        difference_in_seconds = abs((stop_date - start_date).seconds)
        return difference_in_seconds

    def __process_duration_from_api(self, duration):
        processed_duration = None
        try:
            processed_duration = float(duration)
        except Exception:
            match = re.fullmatch(r'(\d+)m(\d+)s', duration)
            if match:
                minutes = float(match.group(1))
                seconds = float(match.group(2))
                total_secs = (60.0 * minutes) + seconds
                processed_duration = total_secs
                return processed_duration

            match = re.fullmatch(r'\d+:\d+:\d+(\.\d+)?', duration)
            if match:
                total_secs = sum(
                    [a * b for a, b in zip([3600.0, 60.0, 1.0], map(float, duration.split(':')))])
                processed_duration = total_secs
                return processed_duration

        return processed_duration

    @metadata
    def get_rich_item(self, item):
        eitem = {}

        self.copy_raw_fields(self.RAW_FIELDS_COPY, item, eitem)

        # The real data
        func_test = item['data']

        # data fields to copy
        copy_fields = ["start_date", "stop_date", "case_name", "criteria", "scenario",
                       "version", "pod_name", "installer", "build_tag", "trust_indicator"]
        for f in copy_fields:
            if f in func_test:
                eitem[f] = func_test[f]
            else:
                eitem[f] = None

        # Fields which names are translated
        map_fields = {"_id": "api_id",
                      "project_name": "project"
                      }
        for fn in map_fields:
            if fn in func_test:
                eitem[map_fields[fn]] = func_test[fn]
            else:
                eitem[map_fields[fn]] = None

        if 'details' in func_test and func_test['details']:
            if 'tests' in func_test['details']:
                if isinstance(func_test['details']['tests'], int):
                    # Only propagate tests if it is a number
                    eitem['tests'] = func_test['details']['tests']
            if 'failures' in func_test['details']:

                if type(func_test['details']['failures']) == list:
                    eitem['failures'] = len(func_test['details']['failures'])
                else:
                    eitem['failures'] = func_test['details']['failures']

            if 'duration' in func_test['details']:
                eitem['duration_from_api'] = self.__process_duration_from_api(func_test['details']['duration'])

                if eitem['duration_from_api'] is None:
                    logger.debug("[functest] Duration from api {} not processed for enriched item {}".format(
                                 func_test['details']['duration'], eitem))

            if 'start_date' in func_test and 'stop_date' in func_test and func_test['stop_date']:
                start_date = self.__convert_str_to_datetime(func_test['start_date'])
                stop_date = self.__convert_str_to_datetime(func_test['stop_date'])
                eitem['duration'] = self.__process_duration(start_date, stop_date)

                if eitem['duration'] is None:
                    logger.debug("[functest] Duration not calculated for enriched item {} with start_date,"
                                 " stop_date: {}, {}".format(eitem, start_date, stop_date))

        if 'duration_from_api' not in eitem:
            eitem['duration_from_api'] = None
        if 'duration' not in eitem:
            eitem['duration'] = None
        if 'failures' not in eitem:
            eitem['failures'] = None
        if 'tests' not in eitem:
            eitem['tests'] = None

        eitem.update(self.get_grimoire_fields(func_test['start_date'], "func_test"))

        # The project is a field already included in the raw data
        # if self.prjs_map:
        #     eitem.update(self.get_item_project(eitem))

        # Hack to show BOOST_PROJECTS first in the donut vis
        eitem['boost_list'] = ['all']
        if eitem['project'] and eitem['project'].lower() in self.BOOST_PROJECTS:
            eitem['boost_list'] += ['boosted']

        self.add_repository_labels(eitem)
        self.add_metadata_filter_raw(eitem)
        return eitem

    @staticmethod
    def __convert_str_to_datetime(text):
        try:
            str_date = str_to_datetime(text)
        except Exception:
            try:
                str_date = unixtime_to_datetime(text)
            except Exception:
                str_date = None

        return str_date



# -*- coding: utf-8 -*-
#
# Copyright (C) 2015-2020 Bitergia
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
# Authors:
#   Jose Javier Merchante Picazo <jjmerchante@cauldron.io>
#

from grimoire_elk.identities.identities import Identities


class GitHubIdentities(Identities):

    @classmethod
    def anonymize_item(cls, item):
        """Remove or hash the fields that contain personal information"""

        category = item['category']

        item = item['data']
        comments_attr = None
        if category == "issue":
            identity_types = ['user', 'assignee']
            comments_attr = 'comments_data'
        elif category == "pull_request":
            identity_types = ['user', 'merged_by']
            comments_attr = 'review_comments_data'
        else:
            identity_types = []

        for identity in identity_types:
            if identity not in item:
                continue
            if not item[identity]:
                continue

            identity_attr = identity + "_data"

            item[identity] = {
                'login': cls._hash(item[identity]['login'])
            }

            item[identity_attr] = {
                'name': cls._hash(item[identity_attr]['login']),
                'login': cls._hash(item[identity_attr]['login']),
                'email': None,
                'company': None,
                'location': None,
            }

        comments = item.get(comments_attr, [])
        for comment in comments:
            if 'user' in comment and comment['user']:
                comment['user'] = {
                    'login': cls._hash(comment['user']['login'])
                }
            comment['user_data'] = {
                'name': cls._hash(comment['user_data']['login']),
                'login': cls._hash(comment['user_data']['login']),
                'email': None,
                'company': None,
                'location': None,
            }
            for reaction in comment['reactions_data']:
                reaction['user'] = {
                    'login': cls._hash(reaction['user']['login'])
                }
